# -*- coding: utf-8 -*-
"""
Contains any static and global variables for MythTV Python Bindings
"""

MYTHXMLCONN = None
MYTHDBCONN = {}
SCHEMA_VERSION = 1246
MVSCHEMA_VERSION = 1031
PROTO_VERSION = 51
PROGRAM_FIELDS = 47
BACKEND_SEP = '[]:[]'